/*
niced - daemon controlant automatiquement les priorites des processus
lou@bocal.cs.univ-paris8.fr
juin 2005
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>		 // fork
#include <math.h>                // log
#include <syslog.h>              // syslog
#include <signal.h>              // signal
#include <stdarg.h>              // Les listes variables d'arguments

#include "niced.h"
/*****************************************************************************************/

/*pointera sur syslog ou imprimer_sur_terminal*/
void (*envoyer_message)(int,const char *,...);
/*****************************************************************************************/

static void afficher_aide(){
  fprintf(stderr,
	  "niced - daemon controlant automatiquement les priorites des processus\nUSAGE : \n\tniced [-c  coef] [-d] [-e exclude] [-h] [-p p:t]\nOPTIONS :\n\t-h\tAffiche l\'aide.\n\t-c coef\tCoefficient affectant les temps de transition entre deux priorites.\n\t-p p:t\tRegle automatiquement le coefficient pour que la priorite p soit atteinte au bout de t secondes.La valeur par defaut est 19:3600.\n\t-e exclude\texclut les processus listes dans le fichier exclude.\n\t-d\tMode de mise au point.\nFILES :\n\texclude\tFichier contenant une liste de nom de programme  ne  devant  pas etre  pris  en  compte  par  niced.\n");
  return ;
}
/***************************************************************************************/

static double calculer_coef(char * _av1){
  int i;
  int p,t;
  char * _av2;
  int length=strlen(_av1);
  for(i=0;i<length;i++)
    if(_av1[i]==':'){_av1[i]='\0';_av2=&_av1[i+1];break;}
  p = atoi(_av1);
  t = atoi(_av2);
  if (p>0&&t>1)return (log(p+1)/log(t));
  return -1;
}
/**************************************************************************************/

void imprimer_sur_terminal(int UNUSED_VARIABLE,const char * format,...){
  /*DESOLE CETTE FONCTION C'EST DU VRAI COPIER COLLER*/
  /*http://membres.lycos.fr/dancel/c/c70_140.htm*/

  va_list pa;
  int n;
  char *s, c;
  float f;

  va_start(pa, format);
  while (*format != '\0') {
    if ( *format == '%' ) {
      switch (*++format) {
	case '%' : putchar('%'); break;
	case 'c' : /* affichage d'un caractre */
		   c = va_arg(pa, int);
		   putchar(c);
		   break;
	case 'd' : /* affichage d'un entier */
		   n = va_arg(pa, int);
		   printf("%d", n);
		   break;
	case 'f' : /* affichage d'un float */
		   f = va_arg(pa, double);    /* !!!!! */
		   printf("%f", f);
		   break;
	case 's' : /* affichage d'une chane */
		   s = va_arg(pa, char *);
		   for ( ; *s != '\0'; s++ )
		     putchar( *s );
		   break;
      } /* end switch */
    }
    else
      putchar( *format );
    format++;
  }
  va_end(pa);

  /*finalement une petite ligne en plus*/
  /*En parallele syslogd rajoute des '\n'*/
  putchar('\n');

  return;
}
/***************************************************************************************/

static void lancer_daemon (char * bin,int mode)
	{
	pid_t pid_ret_fork = fork ();

	if (pid_ret_fork < 0){

	  fprintf(stderr,"Echec de fork ()");
	  exit(1);
	}
	else if (pid_ret_fork > 0)
		{
		// processus pere
		  fprintf (stderr,
			   "%s(%d) vient de creer le fils %d par fork() \nle daemon niced est lance\n",
			   bin, getpid (), pid_ret_fork);
		  closelog();
		  exit (0);
		}
	// processus enfant :
	setsid ();//  lance un programme ds une nouvelle session
	//fermer std in,out,err
	if(mode != MISE_AU_POINT)
	  {
	    fclose(stdin);
	    fclose(stdout);
	    fclose(stderr);
	  }

	chdir ("/");   // changer de dossier courant
	envoyer_message(
		LOG_INFO,
		"niced[%d] a  ete cree par fork() ",
		getpid() );
	}
/***************************************************************************************/

static void traiter_signal (int signal)
	{
	  char signal_ch[7]="SIG****";

	  if (signal == SIGHUP)
	    sprintf(signal_ch,"SIGHUP");
	  else
	    if (signal == SIGINT)
	    sprintf(signal_ch , "SIGINT");
	    else
	      if (signal== SIGQUIT)
		sprintf(signal_ch , "SIGQUIT");
	      else
		if (signal== SIGTERM)
		  sprintf(signal_ch , "SIGTERM");
		else
		  if (signal == SIGABRT)
		    sprintf(signal_ch , "SIGABRT");

	envoyer_message(LOG_INFO,"niced[%d] a recu le signal %s (%d)",getpid (), signal_ch, signal);

	switch (signal)
		{
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:
			envoyer_message (LOG_INFO, "niced[%d] s'arrete par EXIT(0)", getpid ());
			closelog();
			exit (0);
			break;

		case SIGABRT:
			envoyer_message (LOG_INFO, "niced[%d] essaie de s'arreter par ABORT()",getpid ());
			closelog();
			abort();
			break;

		default:
			envoyer_message(LOG_DEBUG, "niced[%d] tourne toujours", getpid ());
		}
	}

/***************************************************************************************/

int main (int argc, char * argv[])
{
  int _ac;

  /*mode daemon par defaut*/
  int mode=0;

  /*valeur par defaut*/
  double coef=(log(19+1)/log(3600));
  double coef_tmp;

  /*definition de la periode du daemon 5mn par defaut*/
  int intervale_inactivite=5;
  int intervale_tmp;

  /*pointeur pour le fichier des excludes*/
  FILE * exclude_fp;
  char * exclu;
  gid_t monGid;
  pid_t monPid;

  /*le deamon utilisara des listes chanees et non des tableaux*/
  List * liste_processus;
  List * liste_excludes=NULL;

  /*Traitement des signaux*/
  signal (SIGHUP, traiter_signal);
  signal (SIGINT, traiter_signal);
  signal (SIGQUIT, traiter_signal);
  signal (SIGTERM, traiter_signal);
  signal (SIGABRT, traiter_signal);

 /*les message seront envoyer par syslogd par defaut*/
  envoyer_message=syslog;

  /*recuperation et traitement des arguments (argv)*/
  for(_ac=1;_ac<argc;_ac++)
    {
      if(!strcmp(argv[_ac],"-h"))
	{
	  afficher_aide();
	  exit(0);
	}
      if(!strcmp(argv[_ac],"-c"))
	{
	  /*fixer un coefficient de calcul de priorite*/
	  coef_tmp=atof(argv[_ac+1]);
	   if(coef_tmp>0)coef=coef_tmp;
	}
      /*OPTION HORS LA MAN PAGE*/
      else if(!strcmp(argv[_ac],"-i"))
	{
	  /*fixer un intervale d'inactivite*/
	  intervale_tmp=atof(argv[_ac+1]);
	  if(intervale_tmp>0) intervale_inactivite=intervale_tmp;
	}
      else if(!strcmp(argv[_ac],"-d"))
	{
	  /*mode de mise au point*/
	  /*les message seront envoyer sur le terminal*/
	  envoyer_message=imprimer_sur_terminal;
	  mode=MISE_AU_POINT;
	}
      else if(!strcmp(argv[_ac],"-e"))
	{
	  /*fichier des excludes*/
	  exclude_fp=fopen(argv[_ac+1],"r");
	  if (exclude_fp!=NULL){
	    liste_excludes=(List *)malloc(sizeof(List));
	    list_init(liste_excludes,free);
	    while(1){
	      exclu=(char *)malloc(sizeof(char)*MAX_CAR_NOM);
	      fgets(exclu,MAX_CAR_NOM,exclude_fp);
	      if(feof(exclude_fp))break;
	      /*suppression de caractere entree*/
	      exclu[strlen(exclu)-1]='\0';
	      list_ins_next(liste_excludes,liste_excludes->queue,(void *)exclu);
	    }
	    fclose(exclude_fp);
	  }
	}
      else
	if(!strcmp(argv[_ac],"-p"))
	  {
	    /*calculer le coef selon p et t*/
	    coef_tmp=calculer_coef(argv[_ac+1]);
	    if(coef_tmp>0)coef=coef_tmp;
	  }
    }

  /*le gid de l'executable*/
  monGid=getgid();

  /*ouvrir le log*/
  openlog ("nice deamon",LOG_NDELAY,LOG_DAEMON);

  /*passer en mode daemon*/
  lancer_daemon(argv[0],mode);

  /*le pid du daemon et non de l'executable*/
  monPid=getpid();

  /*envoyer un message*/
  envoyer_message(
		  LOG_INFO,
		  "niced[%d] demarre, COEF:%f, INACTVITE:%d",
		  monPid,coef,intervale_inactivite);

  /*allocation de memoire pour la liste des processus*/
  liste_processus=(List *)malloc(sizeof(List));

  while (1)
    {
      /*envoyer un message*/
      envoyer_message(
		      LOG_DEBUG,
		      "niced[%d] se reveille",
		     monPid);

      /***CHANGEMENT DE PRIORITES***/
      /*initialiser la liste chainee qui contiendra les processus*/
      list_init(liste_processus,free);
      /*lire le dossier /proc et remplir la liste*/
      remplir_liste_proc(liste_processus,monPid);
      /*changer les priorites*/
      changer_priorites(liste_processus,coef,liste_excludes,monPid,monGid);
      /*detruire la liste des processus*/
      list_destroy(liste_processus);
     /*****************************/

     /*emvoyer un message*/
      envoyer_message(
		      LOG_DEBUG,
		      "niced[%d] s'endort ",
		      monPid);

      /*le deamon s'endore*/
      sleep ( intervale_inactivite );
    }

  return 0;
}
/***************************************************************************************/
