/*  
niced - daemon controlant automatiquement les priorites des processus
lou@bocal.cs.univ-paris8.fr
juin 2005
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>                // log exp
#include <dirent.h>              // dirent et scandir
#include <sys/resource.h>        // setprio
#include <syslog.h>              // syslog

#include "niced.h"
/***************************************************************************************/

static int est_un_exclu(List * liste_excludes,char * tcomm){
  ListElmt *elt;
  int nom_length;
  
  /*ne prendra pas le premier caractere qui est un '('*/
  char * nom=&tcomm[1];
  
  /*ne prendra pas le dernier caractere qui est un ')'*/
  char dernier_c = nom[(nom_length=strlen(nom))-1];
  nom[nom_length-1]='\0';
  
  /*verifier si l'element est dans la liste des excludes*/
  for(elt=list_head(liste_excludes);elt !=NULL; elt = list_next(elt))
    if(!strcmp(list_data(elt),nom))
      {
      nom[nom_length-1]=dernier_c;
      return 1;
      }
  
   /*remettre le dernier caractere*/
  nom[nom_length-1]=dernier_c;
  
  return 0;
}
/***************************************************************************************/

void changer_priorites(List * liste_processus,double coef,List * liste_excludes,pid_t monPid,gid_t monGid){  
  ListElmt *element;//element de la liste de processus
  int ancien_nice,nice;//ancienne et nouvelle priorite
  double nice_double;//nouvelle priorite encore fraiche : exp renvois un double
  double tps;//temps utilisateur + temps systeme
  
  for(element=list_head(liste_processus);element !=NULL; element = list_next(element))
    {//l'element passera d'un processus a l'autre
      if(liste_excludes) 
      {
	if(est_un_exclu(liste_excludes,PS(element)->tcomm))
	  continue;
      }
      ancien_nice=getpriority(PRIO_PROCESS,PS(element)->pid);
      
      /*      calcul de la nouvelle priorite      */
      /**** P = exp(COEF*log(t))-1 ***** log = ln */
      tps = ( PS(element)->utime + PS(element)->stime ) / (double)JIFFIES_PER_SEC;
      nice_double = exp((double)coef*log(tps))-1 ; 
      nice=nice_double;
      
      /*le daemon ne baissera pas les priorites*/
      if(nice<1)
	continue;
      
      /* MAN PAGE : "PAS PLUS DE 19!" */
      if(nice>19)nice=19;
      
      /* Eviter une ecriture*/
      if(ancien_nice == nice)
	continue;
      
      /*l'utilisateur est root*/
      if(monGid==ROOT_GID){
	/*le daemon ne traitera pas les processus du groupe du root*/
	if(PS(element)->pgid==ROOT_GID)
	  continue;
      }
      /*Il fallait juste eviter le cas USER->ROOT PROCESSUS->ROOT*/
      if (! setpriority(
			PRIO_PROCESS, //changement de priorite pour un processus
			((struct process_stat *)list_data(element))->pid, //le processus
			nice)) 
	
	envoyer_message(LOG_INFO,"niced[%d] Changement de priotite du Pid:%d Nom:%s Ancien nice %d -> Nice %d",
			   monPid,PS(element)->pid, PS(element)->tcomm, ancien_nice, nice);
    }
 return;
}
/*****************************************************************************************/

void remplir_liste_proc(List * liste_processus,pid_t monPid){  
  struct dirent **namelist;//tableau de noms de dossiers 
  int pid;
  int n;//nombre de dossier lu
  struct process_stat * prs;//structure de stat d'un processus
  n = scandir(DIR_PROCESS, &namelist, 0, alphasort);
  if (n < 0)
    envoyer_message(LOG_INFO,"niced[%d] : IMPOSSIBLE DE LIRE DANS DIR_PROCESS ",monPid);
  else {
    while(n--) {//TANT QU'IL RESTE DES DOSSIERS A LIRE
      
      if( (pid=atol(namelist[n]->d_name))>0)
	{//le dossier est un processus
	  prs=process_stat(pid);//extraire les stats
	  if(prs)//si OK
	    list_ins_next(liste_processus,liste_processus->queue,(void *)prs);
	  //rajouter un chenon a la fin de liste : un processus de plus
	}
      free (namelist[n]);//liberer la memoire
    }
    free (namelist);
  }
}
/****************************************************************************************/

struct process_stat * process_stat(int pid) {
  FILE* f;
  char buf[BLEN]; /* Buffer gnral (pourvu qu'il ne soit pas trop petit) */
  
  struct process_stat * ps;
  ps=(struct process_stat *)malloc(sizeof(struct process_stat));

  snprintf(buf,BLEN,"/proc/%d/stat",pid);
 
  f = fopen(buf,"r");
  if (f==NULL)
    {
      envoyer_message(LOG_INFO,"Impossible d'ouvrir le fichier de statistique du processus '%d'",pid);
      return NULL;
   }
  if (fgets(buf,BLEN,f)==NULL)
    {
    envoyer_message(LOG_INFO,"Impossible de lire le fichier de statistique du processus '%d'",pid);
    return NULL;
    }
  
  /* C'est fou mais je n'ai pas mieux */
  sscanf(buf,"%d %s %c %d %d %d %d %d %lu %lu \
%lu %lu %lu %d %d %d %d",
	 &ps->pid,
	 (char *)&ps->tcomm, //histoire de mettre gcc -Wall
	 &ps->state,
	 &ps->ppid,
	 &ps->pgid,
	 &ps->sid,
	 &ps->tty_nr,
	 &ps->tty_pgrp,
	 &ps->task_flags,
	 &ps->min_flt,
	 &ps->cmin_flt,
	 &ps->maj_flt,
	 &ps->cmaj_flt,
	 &ps->utime,
	 &ps->stime,
	 &ps->cutime,
	 &ps->cstime );
  
  fclose(f);
  return ps;
}
/******************************************************************************************/
